﻿/* erikhallander@gmail.com */

package com
{
	import flash.display.DisplayObject;
	import flash.display.Sprite;
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.PixelSnapping;	
	
	import flash.events.Event;
	
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.geom.ColorTransform;
	
	import fl.transitions.easing.Strong;
	import flash.filters.BlurFilter;
	
	final public class PixelExploder extends Sprite
	{
		private var orgClip:DisplayObject;
		private var orgBMD:BitmapData;
		
		private var baseColor:uint;
		private var pixelSize:Number;
		
		private var pixelOne:PixelParticle;
		
		private var mHeight:Number;
		private var mWidth:Number;
		
		private var bmpd:BitmapData;
		private var bmp:Bitmap;
		private var bmpFR:Rectangle;
		
		private var ease:Function;
		
		private var time:Number = 0;
		private var dur:Number = 30;
		private var offsetX:Number;
		private var offsetY:Number;
		
		private var scratchPoint:Point = new Point();
		
		private var blur:BlurFilter = new BlurFilter(5, 5, 2);
		private var aTrans:ColorTransform = new ColorTransform();
		
		public function PixelExploder(_orgClip:DisplayObject, aBaseColor:uint = 0x000000, aPixelSize:Number = 2, aWidth:Number = 0, aHeight:Number = 0, transFunc:Function = null)
		{
			baseColor = aBaseColor;
			pixelSize = aPixelSize;
			orgClip = _orgClip;
			
			mHeight = aHeight;
			mWidth = aWidth;
			
			x = orgClip.x;
			y = orgClip.y;
			mouseChildren = false;
			mouseEnabled = false;
			
			if (transFunc === null)
				ease = Strong.easeOut;
			else
				ease = transFunc;
			
			aTrans.alphaMultiplier = .75;
			
			setupPixels();
		}
		private function setupPixels():void
		{
			var h:int = (mHeight > 0) ? mHeight : orgClip.height;
			var w:int = (mWidth > 0) ? mWidth : orgClip.width;
			
			orgBMD = new BitmapData(w, h, true, 0);
			orgBMD.draw(orgClip);
			
			// :TODO: Add a property to set opaque or not (use transparent by default). Perf boost with opaque.
			// :TODO: Add a property for the fill color.
			// :TODO: Base the size increase (overflow) on the property for base radius.
			bmpd = new BitmapData(w*2, h*3, true, baseColor);
			// :TODO: Add properties for PixelSnapping and smoothing as well.
			bmp = new Bitmap(bmpd, PixelSnapping.ALWAYS, false);
			
			var cPixel:PixelParticle;
			
			// :TODO: Base the size expansion on the property for base radius.
			offsetX = w * .5;
			offsetY = h;
			
			for(var i:Number = 0; i<h; i+=pixelSize)
			{
				for(var j:Number = 0; j<w; j+=pixelSize)
				{
					// :TODO: Expand this out to test all pixels, if any are opaque, and not equiv 
					// to the baseColor, the whole rect should be used.
					// :NOTE: this is a performance boost - more pixels = more slow.
					//trace(orgBMD.getPixel(j,i).toString(16) +" == "+ baseColor.toString(16));
					//if (orgBMD.getPixel(j,i) == baseColor || orgBMD.getPixel32(j,i) >> 24 & 0xFF == 0x00) {
					
					// getPixel seems to always return 0x000000 for any pixel with alpha 0
					if (orgBMD.getPixel(j,i) == 0 || orgBMD.getPixel32(j,i) >> 24 & 0xFF == 0x00) {
						continue;
					}
					
					var angle:Number = Math.random() * 360;
					// :TODO: Add a property for the base value of Radius (80).
					var radius:Number = Math.random() * 80;
					
					if (cPixel !== null) {
						cPixel.next = new PixelParticle();
						cPixel = cPixel.next;
					}
					else {
						cPixel = pixelOne = new PixelParticle();
					}
					
					cPixel.bmpRect = new Rectangle(j, i, pixelSize, pixelSize);
					cPixel.origX = cPixel.x = j + offsetX;
					cPixel.origY = cPixel.y = i + offsetY;
					cPixel.tox = Math.round(j+Math.cos(angle)*radius)+ offsetX;
					cPixel.toy = Math.round(i+Math.sin(angle)*radius)+ offsetY;
					
					cPixel.changeX = cPixel.tox - cPixel.origX;
					cPixel.changeY = cPixel.toy - cPixel.origY;
					
					//bmpd.copyPixels(orgBMD, cPixel.bmpRect, new Point(cPixel.x, cPixel.y));
				}
			}
			
			bmpFR = new Rectangle(0, 0, bmp.width, bmp.height);
			//bmpEffect = new Bitmap(bmpd, PixelSnapping.ALWAYS, false);
			//bmpEffect.filters = [blur];
			/*bmpEffectDupe = new Bitmap(bmpd, PixelSnapping.ALWAYS, false);
			bmpEffectDupe.filters = [blur];*/
			bmp.x = -offsetX;
			bmp.y = -offsetY;
			//bmpEffectDupe.x = 
			//bmpEffectDupe.y = 
			
			//addChild(bmpEffectDupe)
			visible = false;
			/*bmp.x = -offsetX;bmp.y = -offsetY;*/
			addChild(bmp);
		}
		
		public function explode(aDur:Number = .7):void
		{
			if (0 < aDur) {
				dur = stage.frameRate * aDur;
				orgClip.visible = false;
				visible = true;
				//alpha = 1;
				time = 5;
				
				// Sets up initial bitmap copy of orgBMD to explode.
				// Also sets change props to correct direction.
				var cPixel:PixelParticle = pixelOne;
				bmpd.lock();
				do {
					// Sets change props to correct order.
					cPixel.changeX = cPixel.tox - cPixel.origX;
					cPixel.changeY = cPixel.toy - cPixel.origY;
					scratchPoint.x = cPixel.origX;
					scratchPoint.y = cPixel.origY;
					bmpd.copyPixels(orgBMD, cPixel.bmpRect, scratchPoint);
					cPixel = cPixel.next;
				}
				while (cPixel !== null);
				bmpd.unlock();
				removeEventListener(Event.ENTER_FRAME, onRetractUpdate);
				removeEventListener(Event.ENTER_FRAME, onExplodeUpdate);
				addEventListener(Event.ENTER_FRAME, onExplodeUpdate);
			}
			else {
				//alpha = 0;
				visible = false;
				orgClip.visible = true;
				finExplode();
			}
		}
		
		private function onExplodeUpdate(event:Event):void
		{
			var cPixel:PixelParticle = pixelOne;
			bmpd.lock();
			
			scratchPoint.x = scratchPoint.y = 0;
			bmpd.applyFilter(bmpd, bmpd.rect, scratchPoint, blur);
			bmpd.colorTransform(bmpd.rect, aTrans);
			
			do {
				cPixel.x = ease(time, cPixel.origX, cPixel.changeX, dur);
				cPixel.x = (cPixel.x < cPixel.tox) ? int(cPixel.x)+1 : int(cPixel.x);
				cPixel.y = ease(time, cPixel.origY, cPixel.changeY, dur);
				cPixel.y = (cPixel.y < cPixel.toy) ? int(cPixel.y)+1 : int(cPixel.y);
				scratchPoint.x = cPixel.x;
				scratchPoint.y = cPixel.y;
				bmpd.copyPixels(orgBMD, cPixel.bmpRect, scratchPoint);
				cPixel = cPixel.next;
			}
			while (cPixel !== null);
			bmpd.unlock();
			++time;
			
			//alpha = ease(time, 1, -1, dur);
			
			if (time > dur) {
				//alpha = 0;
				finExplode();
			}
		}
		
		private function finExplode():void {
			removeEventListener(Event.ENTER_FRAME, onExplodeUpdate);
			dispatchEvent(new Event("EXPLODE_DONE"));
		}
		
		public function retract(aDur:Number = 1):void
		{
			if (0 < aDur)
			{
				dur = stage.frameRate * aDur;
				orgClip.visible = false;
				visible = true;
				time = 0;
				
				var cPixel:PixelParticle = pixelOne;
				do {
					cPixel.changeX = cPixel.origX - cPixel.tox;
					cPixel.changeY = cPixel.origY - cPixel.toy;
					cPixel = cPixel.next;
				}
				while (cPixel !== null);
				
				removeEventListener(Event.ENTER_FRAME, onRetractUpdate);
				removeEventListener(Event.ENTER_FRAME, onExplodeUpdate);
				if (!hasEventListener(Event.ENTER_FRAME))
					addEventListener(Event.ENTER_FRAME, onRetractUpdate);
			}
			else {
				finRetract();
			}
		}
		
		private function onRetractUpdate(event:Event):void
		{
			var cPixel:PixelParticle;
			++time;
			if (time < dur * .7)
			{
				bmpd.lock();
				//alpha = ease(time, 0, 1, dur);
				cPixel = pixelOne;
				
				scratchPoint.x = scratchPoint.y = 0;
				bmpd.applyFilter(bmpd, bmpd.rect, scratchPoint, blur);
				bmpd.colorTransform(bmpd.rect, aTrans);
				
				do {
					cPixel.x = Strong.easeOut(time, cPixel.tox, cPixel.changeX, dur);
					cPixel.x = (cPixel.x < cPixel.origX) ? int(cPixel.x)+1 : int(cPixel.x);
					cPixel.y = Strong.easeOut(time, cPixel.toy, cPixel.changeY, dur);
					cPixel.y = (cPixel.y < cPixel.origY) ? int(cPixel.y)+1 : int(cPixel.y);
					scratchPoint.x = cPixel.x;
					scratchPoint.y = cPixel.y;
					bmpd.copyPixels(orgBMD, cPixel.bmpRect, scratchPoint);
					cPixel = cPixel.next;
				}
				while (cPixel !== null);
				bmpd.unlock();
			}
			else {
				//alpha = 1;
				finRetract();
			}
		}
		
		private function finRetract():void
		{
			removeEventListener(Event.ENTER_FRAME, onRetractUpdate);
			visible = false;
			orgClip.visible = true;
			dispatchEvent(new Event("RETRACT_DONE"));
		}
		
	}
}